@@ -53,6 +53,7 @@
 
 
 
+#include <string.h>
 #include <signal.h>
 #include <libguile.h>
 
@@ -103,42 +104,65 @@ inner_main ()
 "(use-modules (mcron config))"
 "(if config-debug (begin (debug-enable 'debug)"
 "                        (debug-enable 'backtrace)))"
-"(use-modules (ice-9 regex))"
+"(use-modules (ice-9 regex) (ice-9 rdelim))"
 "(define command-name (match:substring (regexp-exec (make-regexp \"[[:alpha:]]*$\")"
 "                                                   (car (command-line)))))"
+"(define (mcron-error exit-code . rest)"
+"  (with-output-to-port (current-error-port)"
+"    (lambda ()"
+"      (for-each display (append (list command-name \": \") rest))"
+"      (newline)))"
+"  (if (and exit-code (not (eq? exit-code 0)))"
+"      (primitive-exit exit-code)))"
+"(defmacro catch-mcron-error (. body)"
+"  `(catch 'mcron-error"
+"          (lambda ()"
+"            ,@body)"
+"          (lambda (key exit-code . msg)"
+"            (apply mcron-error exit-code msg))))"
 "(define command-type (cond ((string=? command-name \"mcron\") 'mcron)"
 "                           ((or (string=? command-name \"cron\")"
 "                                (string=? command-name \"crond\")) 'cron)"
 "                           ((string=? command-name \"crontab\") 'crontab)"
 "                           (else"
-"                            (display \"The command name is invalid.\\n\")"
-"                            (primitive-exit 12))))"
+"                            (mcron-error 12 \"The command name is invalid.\"))))"
 "(use-modules (ice-9 getopt-long))"
 "(define options"
-"  (getopt-long (command-line)"
-"               (append"
-"                (case command-type ('crontab"
-"                            '((user     (single-char #\\u) (value #t))"
-"                              (edit     (single-char #\\e) (value #f))"
-"                              (list     (single-char #\\l) (value #f))"
-"                              (remove   (single-char #\\r) (value #f))))"
-"                      (else `((schedule (single-char #\\s) (value optional))"
-"                              (daemon   (single-char #\\d) (value #f))"
-"                              (noetc    (single-char #\\n) (value #f))"
-"                              (stdin    (single-char #\\i) (value #t)"
-"                                        (predicate"
-"                                         ,(lambda (value)"
+"  (catch"
+"   'misc-error"
+"   (lambda ()"
+"     (getopt-long (command-line)"
+"                  (append"
+"                   (case command-type"
+"                     ((crontab)"
+"                      '((user     (single-char #\\u) (value #t))"
+"                        (edit     (single-char #\\e) (value #f))"
+"                        (list     (single-char #\\l) (value #f))"
+"                        (remove   (single-char #\\r) (value #f))))"
+"                     (else `((schedule (single-char #\\s)"
+"                                       (value optional)"
+"                                       (predicate "
+"                                        ,(lambda (value)"
+"                                           (or (eq? value #t)"
+"                                               (string->number value)))))"
+"                             (daemon   (single-char #\\d) (value #f))"
+"                             (noetc    (single-char #\\n) (value #f))"
+"                             (stdin    (single-char #\\i) (value #t)"
+"                                       (predicate"
+"                                        ,(lambda (value)"
 "                                           (or (string=? \"vixie\" value)"
 "                                               (string=? \"guile\" value))))))))"
-"                '((version  (single-char #\\v) (value #f))"
-"                  (help     (single-char #\\h) (value #f))))))"
+"                   '((version  (single-char #\\v) (value #f))"
+"                     (help     (single-char #\\h) (value #f))))))"
+"   (lambda (key func fmt args . rest)"
+"     (mcron-error 1 (apply format (append (list #f fmt) args))))))"
 "(if (option-ref options 'version #f)"
 "    (begin"
 "      (display (string-append \"\\n"
 "\" command-name \"  (\" config-package-string \")\\n"
 "Written by Dale Mellor\\n"
 "\\n"
-"Copyright (C) 2003  Dale Mellor\\n"
+"Copyright (C) 2003, 2006  Dale Mellor\\n"
 "This is free software; see the source for copying conditions.  There is NO\\n"
 "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\\n"
 "\"))"
@@ -147,7 +171,8 @@ inner_main ()
 "    (begin"
 "      (display (string-append \""
 "Usage: \" (car (command-line))"
-"(case command-type ('mcron"
+"(case command-type"
+"  ((mcron)"
 "\" [OPTIONS] [FILES]\\n"
 "Run an mcron process according to the specifications in the FILES (`-' for\\n"
 "standard input), or use all the files in ~/.cron with .guile or .vixie\\n"
@@ -159,9 +184,9 @@ inner_main ()
 "                              will be run by mcron\\n"
 "  -d, --daemon              Immediately detach the program from the terminal and\\n"
 "                              run as a daemon process\\n"
-"  -i, --stdin=(guile|vixie) Format of data passed as standard input\\n"
-"                              (default guile)\")"
-"  ('cron"
+"  -i, --stdin=(guile|vixie) Format of data passed as standard input or\\n"
+"                              file arguments (default guile)\")"
+"  ((cron)"
 "\" [OPTIONS]\\n"
 "Unless an option is specified, run a cron daemon as a detached process, \\n"
 "reading all the information in the users' crontabs and in /etc/crontab.\\n"
@@ -172,13 +197,14 @@ inner_main ()
 "                              will be run by cron\\n"
 "  -n, --noetc               Do not check /etc/crontab for updates (HIGHLY\\n"
 "                              RECOMMENDED).\")"
-"  ('crontab"
+"  ((crontab)"
 "           (string-append \" [-u user] file\\n\""
 "           \"       \" (car (command-line)) \" [-u user] { -e | -l | -r }\\n\""
 "           \"               (default operation is replace, per 1003.2)\\n\""
 "           \"       -e      (edit user's crontab)\\n\""
 "           \"       -l      (list user's crontab)\\n\""
-"           \"       -r      (delete user's crontab)\\n\")))"
+"           \"       -r      (delete user's crontab)\\n\"))"
+"  (else \"rubbish\"))"
 "\"\\n\\n"
 "Report bugs to \" config-package-bugreport \".\\n"
 "\"))"
@@ -191,16 +217,16 @@ inner_main ()
 "(if (eq? command-type 'cron)"
 "    (begin"
 "      (if (not (eqv? (getuid) 0))"
-"          (begin"
-"            (display \"This program must be run by the root user (and should \")"
-"            (display \"have been installed as such).\\n\")"
-"            (primitive-exit 16)))"
+"          (mcron-error 16"
+"                       \"This program must be run by the root user (and should \""
+"                       \"have been installed as such).\"))"
 "      (if (access? config-pid-file F_OK)"
-"          (begin"
-"            (display \"A cron daemon is already running.\\n\")"
-"            (display \"  (If you are sure this is not true, remove the file\\n\")"
-"            (display \"   \" config-pid-file \".)\\n\")"
-"            (primitive-exit 1)))"
+"          (mcron-error 1"
+"		       \"A cron daemon is already running.\\n\""
+"		       \"  (If you are sure this is not true, remove the file\\n\""
+"		       \"   \""
+"		       config-pid-file"
+"		       \".)\"))"
 "      (if (not (option-ref options 'schedule #f))"
 "          (with-output-to-file config-pid-file noop))"
 "      (setenv \"MAILTO\" #f)"
@@ -235,22 +261,28 @@ inner_main ()
 "(define crontab-real-user (passwd:name (getpw (getuid))))"
 "(if (or (eq? (in-access-file? config-allow-file crontab-real-user) #f)"
 "        (eq? (in-access-file? config-deny-file crontab-real-user) #t))"
-"    (begin"
-"      (display \"Access denied by system operator.\\n\")"
-"      (primitive-exit 6)))"
+"    (mcron-error 6 \"Access denied by system operator.\"))"
 "(if (> (+ (if (option-ref options 'edit #f) 1 0)"
 "          (if (option-ref options 'list #f) 1 0)"
 "          (if (option-ref options 'remove #f) 1 0))"
 "       1)"
-"    (begin"
-"      (display \"crontab: Only one of options -e, -l or -r can be used.\\n\")"
-"      (primitive-exit 7)))"
+"    (mcron-error 7 \"Only one of options -e, -l or -r can be used.\"))"
 "(if (and (not (eqv? (getuid) 0))"
 "         (option-ref options 'user #f))"
-"    (begin (display \"crontab: Only root can use the -u option.\\n\")"
-"           (primitive-exit 8)))"
+"    (mcron-error 8 \"Only root can use the -u option.\"))"
 "(define crontab-user (option-ref options 'user crontab-real-user))"
 "(define crontab-file (string-append config-spool-dir \"/\" crontab-user))"
+"(define (get-yes-no prompt . re-prompt)"
+"  (if (not (null? re-prompt))"
+"      (display \"Please answer y or n.\\n\"))"
+"  (display (string-append prompt \" \"))"
+"  (let ((r (read-line)))"
+"    (if (not (string-null? r))"
+"        (case (string-ref r 0)"
+"          ((#\\y #\\Y) #t)"
+"          ((#\\n #\\N) #f)"
+"          (else (get-yes-no prompt #t)))"
+"        (get-yes-no prompt #t))))"
 "(cond"
 " ((option-ref options 'list #f)"
 "  (catch #t (lambda ()"
@@ -269,10 +301,20 @@ inner_main ()
 "    (catch #t (lambda () (copy-file crontab-file temp-file))"
 "              (lambda (key . args) (with-output-to-file temp-file noop)))"
 "    (chown temp-file (getuid) (getgid))"
-"    (system (string-append (or (getenv \"VISUAL\") (getenv \"EDITOR\") \"vi\")"
-"                           \" \""
-"                           temp-file))"
-"    (read-vixie-file temp-file)"
+"    (let retry ()"
+"      (system (string-append"
+"               (or (getenv \"VISUAL\") (getenv \"EDITOR\") \"vi\")"
+"               \" \""
+"               temp-file))"
+"      (catch 'mcron-error"
+"             (lambda () (read-vixie-file temp-file))"
+"             (lambda (key exit-code . msg)"
+"               (apply mcron-error 0 msg)"
+"               (if (get-yes-no \"Edit again?\")"
+"                   (retry)"
+"                   (begin"
+"                     (mcron-error 0 \"Crontab not changed\")"
+"                     (primitive-exit 0))))))"
 "    (copy-file temp-file crontab-file)"
 "    (delete-file temp-file)"
 "    (hit-server crontab-user)))"
@@ -282,19 +324,26 @@ inner_main ()
 "            noop))"
 " ((not (null? (option-ref options '() '())))"
 "  (let ((input-file (car (option-ref options '() '()))))"
-"    (if (string=? input-file \"-\")"
-"        (let ((input-string (stdin->string)))"
-"          (read-vixie-port (open-input-string input-string))"
-"          (with-output-to-file crontab-file (lambda ()"
-"                                              (display input-string))))"
-"        (begin"
-"          (read-vixie-file input-file)"
-"          (copy-file input-file crontab-file))))"
-"  (hit-server crontab-user))"
+"    (catch-mcron-error"
+"     (if (string=? input-file \"-\")"
+"         (let ((input-string (stdin->string)))"
+"           (read-vixie-port (open-input-string input-string))"
+"           (with-output-to-file crontab-file (lambda ()"
+"                                               (display input-string))))"
+"         (begin"
+"           (read-vixie-file input-file)"
+"           (copy-file input-file crontab-file))))"
+"    (hit-server crontab-user)))"
 " (else"
-"  (display \"crontab: usage error: file name must be specified for replace.\\n\")"
-"  (primitive-exit 15)))"
+"  (mcron-error 15 \"usage error: file name must be specified for replace.\")))"
 "      (quit)))"
+"(define (regular-file? file)"
+"  (catch 'system-error"
+"	 (lambda ()"
+"	   (eq? (stat:type (stat file)) 'regular))"
+"	 (lambda (key call fmt args . rest)"
+"	   (mcron-error 0 (apply format (append (list #f fmt) args)))"
+"	   #f)))"
 "(define guile-file-regexp (make-regexp \"\\\\.gui(le)?$\"))"
 "(define vixie-file-regexp (make-regexp \"\\\\.vix(ie)?$\"))"
 "(define (process-user-file file-path)"
@@ -313,12 +362,11 @@ inner_main ()
 "                     (directory (opendir dir-path)))"
 "                (do ((file-name (readdir directory) (readdir directory)))"
 "                    ((eof-object? file-name) (closedir directory))"
-"                    (process-user-file (string-append dir-path"
-"                                                      \"/\""
-"                                                      file-name)))))"
+"                  (process-user-file (string-append dir-path"
+"                                                    \"/\""
+"                                                    file-name)))))"
 "         (lambda (key . args)"
-"           (display \"Cannot read files in your ~/.cron directory.\\n\")"
-"           (primitive-exit 13))))"
+"           (mcron-error 13 \"Cannot read files in your ~/.cron directory.\"))))"
 "(define (valid-user user-name)"
 "  (setpwent)"
 "  (do ((entry (getpw) (getpw)))"
@@ -328,39 +376,43 @@ inner_main ()
 "       entry)))"
 "(use-modules (srfi srfi-2)) "
 "(define (process-files-in-system-directory)"
-"  (catch #t (lambda ()"
-"              (let ((directory (opendir config-spool-dir)))"
-"                (do ((file-name (readdir directory) (readdir directory)))"
-"                    ((eof-object? file-name))"
-"                  (and-let* ((user (valid-user file-name)))"
-"                            (set-configuration-user user) "
-"                            (read-vixie-file (string-append config-spool-dir"
-"                                                            \"/\""
-"                                                            file-name))))))"
-"      (lambda (key . args)"
-"        (display \"You do not have permission to access the system crontabs.\\n\")"
-"        (primitive-exit 4))))"
+"  (catch #t"
+"         (lambda ()"
+"           (let ((directory (opendir config-spool-dir)))"
+"             (do ((file-name (readdir directory) (readdir directory)))"
+"                 ((eof-object? file-name))"
+"               (and-let* ((user (valid-user file-name)))"
+"                         (set-configuration-user user) "
+"                         (catch-mcron-error"
+"                          (read-vixie-file (string-append config-spool-dir"
+"                                                          \"/\""
+"                                                          file-name)))))))"
+"         (lambda (key . args)"
+"           (mcron-error"
+"            4"
+"            \"You do not have permission to access the system crontabs.\"))))"
 "(case command-type"
-"  ('mcron (if (null? (option-ref options '() '()))"
-"              (process-files-in-user-directory)"
-"              (for-each (lambda (file-path)"
-"                          (process-user-file file-path))"
-"                        (option-ref options '() '()))))"
-"  ('cron (process-files-in-system-directory)"
-"         (use-system-job-list)"
-"         (read-vixie-file \"/etc/crontab\" parse-system-vixie-line)"
-"         (use-user-job-list)"
-"         (if (not (option-ref options 'noetc #f))"
-"             (begin"
-"               (display"
-"\"WARNING: cron will check for updates to /etc/crontab EVERY MINUTE. If you do\\n"
+"  ((mcron) (if (null? (option-ref options '() '()))"
+"                (process-files-in-user-directory)"
+"                (for-each (lambda (file-path)"
+"                            (process-user-file file-path))"
+"                          (option-ref options '() '()))))"
+"  ((cron) (process-files-in-system-directory)"
+"   (use-system-job-list)"
+"   (catch-mcron-error"
+"    (read-vixie-file \"/etc/crontab\" parse-system-vixie-line))"
+"   (use-user-job-list)"
+"   (if (not (option-ref options 'noetc #f))"
+"       (begin"
+"         (display"
+"          \"WARNING: cron will check for updates to /etc/crontab EVERY MINUTE. If you do\\n"
 "not use this file, or you are prepared to manually restart cron whenever you\\n"
 "make a change, then it is HIGHLY RECOMMENDED that you use the --noetc\\n"
 "option.\\n\")"
-"               (set-configuration-user \"root\")"
-"               (job '(- (next-minute-from (next-minute)) 6)"
-"                    check-system-crontab"
-"                    \"/etc/crontab update checker.\")))))"
+"         (set-configuration-user \"root\")"
+"         (job '(- (next-minute-from (next-minute)) 6)"
+"              check-system-crontab"
+"              \"/etc/crontab update checker.\")))))"
 "(and-let* ((count (option-ref options 'schedule #f)))"
 "          (set! count (if (eq? count #t)"
 "                          8"
@@ -377,28 +429,39 @@ inner_main ()
 "            (lambda () (display (getpid)) (newline))))))"
 "(define fd-list '())"
 "(if (eq? command-type 'cron)"
-"    (let ((socket (socket AF_UNIX SOCK_STREAM 0)))"
-"      (bind socket AF_UNIX config-socket-file)"
-"      (listen socket 5)"
-"      (set! fd-list (list socket))))"
+"    (catch #t"
+"           (lambda ()"
+"             (let ((socket (socket AF_UNIX SOCK_STREAM 0)))"
+"               (bind socket AF_UNIX config-socket-file)"
+"               (listen socket 5)"
+"               (set! fd-list (list socket))))"
+"           (lambda (key . args)"
+"             (delete-file config-pid-file)"
+"             (mcron-error 1"
+"                          \"Cannot bind to UNIX socket \""
+"                          config-socket-file))))"
+"		     "
 "(define (process-update-request)"
 "  (let* ((socket (car (accept (car fd-list))))"
 "         (user-name (read-line socket)))"
 "    (close socket)"
 "    (set-configuration-time (current-time))"
-"    (if (string=? user-name \"/etc/crontab\")"
-"        (begin"
-"          (clear-system-jobs)"
-"          (use-system-job-list)"
-"          (read-vixie-file \"/etc/crontab\" parse-system-vixie-line)"
-"          (use-user-job-list))"
-"        (let ((user (getpw user-name)))"
-"          (remove-user-jobs user)"
-"          (set-configuration-user user)"
-"          (read-vixie-file (string-append config-spool-dir \"/\" user-name))))))"
-"(while #t"
-"       (run-job-loop fd-list)"
-"       (process-update-request))"
+"    (catch-mcron-error"
+"     (if (string=? user-name \"/etc/crontab\")"
+"         (begin"
+"           (clear-system-jobs)"
+"           (use-system-job-list)"
+"           (read-vixie-file \"/etc/crontab\" parse-system-vixie-line)"
+"           (use-user-job-list))"
+"         (let ((user (getpw user-name)))"
+"           (remove-user-jobs user)"
+"           (set-configuration-user user)"
+"           (read-vixie-file (string-append config-spool-dir \"/\" user-name)))))))"
+"(catch-mcron-error"
+" (while #t"
+"        (run-job-loop fd-list)"
+"        (if (not (null? fd-list))"
+"            (process-update-request))))"
                                                           ) );
 }
 
